{% extends 'sos-full.html.j2' %}

{% import 'parts/control_panel.tpl' as control_panel %}

{% block header %}
{{ super() }}
{{ control_panel.css() }}
{% endblock header %}

{%- block input -%}
  {%- if 'scratch' in cell.metadata.get('tags', []) -%}
	{%- elif 'report_cell' in cell.metadata.get('tags', []) -%}
        {{ super() }}
  {%- else -%}
        <div class="hidden_content">
        {{ super() }}
        </div>
  {%- endif -%}
{%- endblock input -%}

{% block output %}
  {%- if 'report_output' in cell.metadata.get('tags', []) -%}
      {{ super() }}
  {%- elif 'report_cell' in cell.metadata.get('tags', []) -%}
      {{ super() }}
  {%- elif 'scratch' in cell.metadata.get('tags', []) -%}
  {%- else -%}
      <div class="hidden_content">
      {{ super() }}
      </div>
  {%- endif -%}
{% endblock output %}

{% block markdowncell %}
  {%- if 'hide_output' in cell.metadata.get('tags', []) -%}
	<div class="hidden_content">
      {{ super() }}
	</div>
  {%- elif 'scratch' in cell.metadata.get('tags', []) -%}
  {%- else -%}
      {{ super() }}
  {%- endif -%}
{%- endblock markdowncell -%}


{% block body %}
{{ control_panel.html() }}
{{ super() }}
{% endblock body %}

{% block footer %}
{{ control_panel.js() }}
{{ super() }}
{% endblock footer %}
